from collections import namedtuple

import numpy as np

from nptyping import Array

EvaluationMetrics = namedtuple(
    "EvaluationMetrics",
    ["tp", "fp", "fn", "precision", "recall", "f1"]
)

def compute_metrics(
    labels: Array[np.int32],
    preds: Array[np.int32]
) -> EvaluationMetrics:

    # manually implemented due to keep no. of dependencies minimal
    tp = np.sum(preds * labels)
    fp = np.sum(preds * (1-labels))
    fn = np.sum((1-preds) * labels)

    precision = tp / (tp+fp)
    recall = tp / (tp+fn)
    f1 = 2 * precision * recall / (precision + recall)

    return EvaluationMetrics(
        tp=tp,
        fp=fp,
        fn=fn,
        precision=precision,
        recall=recall,
        f1=f1
    )



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from attacut import utils, dataloaders, logger
from . import BaseModel, ConvolutionBatchNorm

log = logger.get_logger(__name__)


class Model(BaseModel):
    dataset = dataloaders.SyllableCharacterSeqDataset

    def __init__(self, data_config, model_config="embc:16|embs:8|conv:16|l1:16|do:0.0"):
        super(Model, self).__init__()

        no_chars = data_config['num_char_tokens']
        log.info("no. characters: %d" % no_chars)

        no_syllables = data_config['num_tokens']
        log.info("no. syllables: %d" % no_syllables)

        config = utils.parse_model_params(model_config)
        conv_filters = config['conv']
        dropout_rate = config.get("do", 0)

        self.ch_embeddings = nn.Embedding(
            no_chars,
            config['embc'],
            padding_idx=0
        )

        self.sy_embeddings = nn.Embedding(
            no_syllables,
            config['embs'],
            padding_idx=0
        )

        emb_dim = config['embc'] + config['embs']

        self.dropout= torch.nn.Dropout(p=dropout_rate)

        self.conv1 = ConvolutionBatchNorm(emb_dim, conv_filters, 3)
        self.conv2 = ConvolutionBatchNorm(emb_dim, conv_filters, 5, dilation=3)
        self.conv3 = ConvolutionBatchNorm(emb_dim, conv_filters, 9, dilation=2)

        self.linear1 = nn.Linear(conv_filters, config['l1'])
        self.linear2 = nn.Linear(config['l1'], 1)

        self.model_params = model_config

    def forward(self, inputs):
        x, seq_lengths = inputs

        x_char, x_syllable = x[:, 0, :], x[:, 1, :]

        ch_embedding = self.ch_embeddings(x_char)
        sy_embedding = self.sy_embeddings(x_syllable)

        embedding = torch.cat((ch_embedding, sy_embedding), dim=2)

        embedding = embedding.permute(0, 2, 1)

        conv1 = self.dropout(self.conv1(embedding).permute(0, 2, 1))
        conv2 = self.dropout(self.conv2(embedding).permute(0, 2, 1))
        conv3 = self.dropout(self.conv3(embedding).permute(0, 2, 1))

        out = torch.stack((conv1, conv2, conv3), 3)

        out, _ = torch.max(out, 3)

        out = F.relu(self.linear1(out))
        out = self.linear2(out)

        out = out.view(-1)

        return out


from typing import Dict, List

import torch

from attacut import (artifacts, dataloaders, logger, models, preprocessing,
                     utils)

log = logger.get_logger(__name__)


def tokenize(txt: str) -> List[str]:
    return SingletonTokenizer().tokenize(txt)


class Tokenizer:
    def __init__(self, model: str = "attacut-sc"):
        # resolve model's path
        model_path = artifacts.get_path(model)

        params = utils.load_training_params(model_path)

        model_name = params.name
        log.info("loading model %s" % model_name)

        model_cls: models.BaseModel = models.get_model(model_name)

        # instantiate dataset
        dataset: dataloaders.SequenceDataset = model_cls.dataset()

        # load necessary dicts into memory
        data_config: Dict = dataset.setup_featurizer(model_path)

        # instantiate model
        self.model = model_cls.load(
            model_path,
            data_config,
            params.params
        )

        self.dataset = dataset

    def tokenize(self, txt: str, sep="|", device="cpu", pred_threshold=0.5) -> List[str]:
        tokens, features = self.dataset.make_feature(txt)

        inputs = (
            features,
            torch.Tensor(0)  # dummy label when won't need it here
        )

        x, _, _ = self.dataset.prepare_model_inputs(inputs, device=device)
        logits = torch.sigmoid(self.model(x))

        preds = logits.cpu().detach().numpy() > pred_threshold

        words = preprocessing.find_words_from_preds(tokens, preds)

        return words


class SingletonTokenizer(Tokenizer):
    _instance = None
    _total_object = 0  # for testing, should be not more than `1`

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(SingletonTokenizer, cls)\
                .__new__(cls, *args, **kwargs)

            cls._total_object += 1
        return cls._instance



import importlib
import re

import torch
import torch.nn as nn

import attacut
from attacut import logger

log = logger.get_logger(__name__)


class ConvolutionBatchNorm(nn.Module):
    def __init__(self, channels, filters, kernel_size, stride=1, dilation=1):
        super(ConvolutionBatchNorm, self).__init__()

        padding = kernel_size // 2
        padding += padding * (dilation-1)

        self.conv = nn.Conv1d(
            channels,
            filters,
            kernel_size,
            stride=stride,
            dilation=dilation,
            padding=padding
        )

        self.bn = nn.BatchNorm1d(filters)

    def forward(self, x):
        return self.bn(self.conv(x))

class ConvolutionLayer(nn.Module):
    def __init__(self, channels, filters, kernel_size, stride=1, dilation=1):
        super(ConvolutionLayer, self).__init__()

        padding = kernel_size // 2
        padding += padding * (dilation-1)

        self.conv = nn.Conv1d(
            channels,
            filters,
            kernel_size,
            stride=stride,
            dilation=dilation,
            padding=padding
        )

    def forward(self, x):
        return self.conv(x)

class BaseModel(nn.Module):
    dataset = None
    @classmethod
    def load(cls, path, data_config, model_config, with_eval=True):
        model = cls(data_config, model_config)

        model_path = "%s/model.pth" % path
        model.load_state_dict(torch.load(model_path, map_location="cpu"))

        log.info("loaded: %s|%s (variables %d)" % (
            model_path,
            model_config,
            model.total_trainable_params()
        ))

        if with_eval:
            log.info("setting model to eval mode")
            model.eval()

        return model

    def total_trainable_params(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)


def get_model(model_name) -> BaseModel:
    module_path = "attacut.models.%s" % model_name
    log.info("Taking %s" % module_path)

    model_mod = importlib.import_module(module_path)
    return model_mod.Model


import os

from attacut import logger

log = logger.get_logger(__name__)

artifact_dir = os.path.dirname(__file__)


def get_path(name: str) -> str:
    if name in ["attacut-c", "attacut-sc"]:
        return f"{artifact_dir}/{name}"
    else:
        # if name isn't in the list, then it's a custom model
        log.info("model_path: %s" % name)
        return name



import logging
import os

logging.basicConfig(
    format="%(asctime)s %(levelname)-8s [%(name)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=os.getenv("ATTACUT_LOG_LEVEL", "warning").upper()
)


def get_logger(name):
    return logging.getLogger(name)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from attacut import utils, dataloaders
from . import BaseModel, ConvolutionLayer


class Model(BaseModel):
    dataset = dataloaders.CharacterSeqDataset

    def __init__(self, data_config, model_config="emb:32|conv:48|l1:16|do:0.1"):
        super(Model, self).__init__()

        no_chars = data_config['num_tokens']

        config = utils.parse_model_params(model_config)
        emb_dim = config['emb']
        conv_filters = config['conv']
        dropout_rate = config.get("do", 0)

        self.embeddings = nn.Embedding(
            no_chars,
            emb_dim,
            padding_idx=0
        )

        self.dropout= torch.nn.Dropout(p=dropout_rate)

        self.conv1 = ConvolutionLayer(emb_dim, conv_filters, 3)
        self.conv2 = ConvolutionLayer(emb_dim, conv_filters, 5, dilation=3)
        self.conv3 = ConvolutionLayer(emb_dim, conv_filters, 9, dilation=2)

        self.linear1 = nn.Linear(conv_filters, config['l1'])
        self.linear2 = nn.Linear(config['l1'], 1)

        self.model_params = model_config

    def forward(self, inputs):
        x, seq_lengths = inputs

        embedding = self.embeddings(x).permute(0, 2, 1)

        conv1 = self.conv1(embedding).permute(0, 2, 1)
        conv2 = self.conv2(embedding).permute(0, 2, 1)
        conv3 = self.conv3(embedding).permute(0, 2, 1)

        out = torch.stack((conv1, conv2, conv3), 3)

        out, _ = torch.max(out, 3)
        out = self.dropout(out)

        out = F.relu(self.linear1(out))
        out = self.linear2(out)

        out = out.view(-1)

        return out


import numpy as np
import torch
from torch.utils.data import Dataset

from attacut import logger, preprocessing, utils

log = logger.get_logger(__name__)


class SequenceDataset(Dataset):
    def __init__(self, path: str = None):
        if path:
            self.load_preprocessed_data(path)

    def __len__(self):
        return self.total_samples

    def __getitem__(self, index):
        return self.data[index]

    def load_preprocessed_data(self, path):
        self.data = []

        suffix = path.split("/")[-1]
        with open(path) as f, \
            utils.Timer("load-seq-data--%s" % suffix) as timer:
            for line in f:
                self.data.append(self._process_line(line))

        self.total_samples = len(self.data)

    def make_feature(self, txt: str):
        raise NotImplementedError

    def setup_featurizer(self, path: str):
        raise NotImplementedError

    @staticmethod
    def prepare_model_inputs(inputs, device="cpu"):

        x, seq_lengths = inputs[0]
        x = x.to(device)
        y = inputs[1].float().to(device).reshape(-1)

        return (x, seq_lengths), y, y.shape[0]

    @staticmethod
    def _process_line(line: str):
        # only use when training
        raise NotImplementedError

    @staticmethod
    def collate_fn(batch):
        # only use when training
        raise NotImplementedError

    @classmethod
    def load_preprocessed_file_with_suffix(cls, dir: str, suffix: str) -> "SequenceDataset":
        path = "%s/%s" % (dir, suffix)
        log.info("Loading preprocessed data from %s" % path)
        return cls(path=path)


class CharacterSeqDataset(SequenceDataset):
    def setup_featurizer(self, path: str):
        self.dict = utils.load_dict(f"{path}/characters.json")

        return dict(num_tokens=len(self.dict))

    def make_feature(self, txt):
        characters = list(txt)
        ch_ix = list(
            map(
                lambda c: preprocessing.character2ix(self.dict, c),
                characters
            )
        )

        features = np.array(ch_ix, dtype=np.int64).reshape((1, -1))

        seq_lengths = np.array([features.shape[-1]], dtype=np.int64)

        return characters, (torch.from_numpy(features), torch.from_numpy(seq_lengths))

    @staticmethod
    def _process_line(line):
        label, indices = line.split("::")

        y = np.array(list(label)).astype(int)
        x = np.array(indices.split(" ")).astype(int)

        seq = len(y)

        return (x, seq), y

    @staticmethod
    def collate_fn(batch):
        total_samples = len(batch)

        seq_lengths = np.array(list(map(lambda x: x[0][1], batch)))
        max_length = np.max(seq_lengths)

        features = np.zeros((total_samples, max_length), dtype=np.int64)
        labels = np.zeros((total_samples, max_length), dtype=np.int64)

        for i, s in enumerate(batch):
            b_feature = s[0][0]
            total_features = len(b_feature)
            features[i, :total_features] = b_feature
            labels[i, :total_features] = s[1]

        seq_lengths = torch.from_numpy(seq_lengths)
        seq_lengths, perm_idx = seq_lengths.sort(0, descending=True)

        inputs = (torch.from_numpy(features)[perm_idx], seq_lengths)

        labels = torch.from_numpy(labels)[perm_idx]

        return inputs, labels


class SyllableCharacterSeqDataset(SequenceDataset):
    def setup_featurizer(self, path: str):
        self.ch_dict = utils.load_dict(f"{path}/characters.json")
        self.sy_dict = utils.load_dict(f"{path}/syllables.json")

        return dict(
            num_char_tokens=len(self.ch_dict),
            num_tokens=len(self.sy_dict)
        )

    def make_feature(self, txt):
        syllables = preprocessing.syllable_tokenize(txt)

        sy2ix, ch2ix = self.sy_dict, self.ch_dict

        ch_ix, syllable_ix = [], []

        for syllable in syllables:
            six = preprocessing.syllable2ix(sy2ix, syllable)

            chs = list(
                map(
                    lambda ch: preprocessing.character2ix(ch2ix, ch),
                    list(syllable)
                )
            )

            ch_ix.extend(chs)
            syllable_ix.extend([six]*len(chs))

        features = np.stack((ch_ix, syllable_ix), axis=0) \
            .reshape((1, 2, -1)) \
            .astype(np.int64)

        seq_lengths = np.array([features.shape[-1]], dtype=np.int64)

        return list(txt), (torch.from_numpy(features), torch.from_numpy(seq_lengths))

    @staticmethod
    def _process_line(line):
        label, ch_indices, sy_indices = line.split("::")

        y = np.array(list(label)).astype(int)

        cx = np.array(ch_indices.split(" ")).astype(int)
        sx = np.array(sy_indices.split(" ")).astype(int)
        x = np.stack((cx, sx), axis=0)

        seq = len(y)

        return (x, seq), y

    @staticmethod
    def collate_fn(batch):
        total_samples = len(batch)

        seq_lengths = np.array(list(map(lambda x: x[0][1], batch)))
        max_length = np.max(seq_lengths)

        features = np.zeros((total_samples, 2, max_length), dtype=np.int64)
        labels = np.zeros((total_samples, max_length), dtype=np.int64)

        for i, s in enumerate(batch):
            b_feature = s[0][0]
            total_features = b_feature.shape[1]
            features[i, :, :total_features] = b_feature
            labels[i, :total_features] = s[1]

        seq_lengths = torch.from_numpy(seq_lengths)
        seq_lengths, perm_idx = seq_lengths.sort(0, descending=True)

        inputs = (torch.from_numpy(features)[perm_idx], seq_lengths)

        labels = torch.from_numpy(labels)[perm_idx]

        return inputs, labels



import re
import string
from typing import Dict, List

import ssg
from attacut.minpythainlp import thai_digit_to_arabic_digit

ARABIC_RX = re.compile(r"[A-Za-z]+")
CAMEL_CASE_RX = re.compile(r"([a-z])([A-Z])([a-z])")
EMAIL_RX = re.compile(r"^\w+\@\w+\.\w+$")
NUMBER_RX = re.compile(r"[0-9,]+")
TRAILING_SPACE_RX = re.compile(r"\n$")
URL_RX = re.compile(r"(https?:\/\/)?(\w+\.)?\w+\.\w+")

DEFAULT_PREPROCESSING_STEPS = [
    "remove_tags", 
    "thai_digit_to_arabic_digit",
    "new_line_as_space",
    "remove_first_pipe",
    "remove_last_pipe"
]


def syllable2token(syllable: str) -> str:
    if ARABIC_RX.match(syllable):
        return "<ENGLISH>"
    elif NUMBER_RX.match(syllable):
        return "<NUMBER>"
    else:
        return syllable


def syllable2ix(sy2ix: Dict[str, int], syllable: str) -> int:
    token = syllable2token(syllable)

    return sy2ix.get(token, sy2ix.get("<UNK>"))


def character2ix(ch2ix: Dict[str, int], character: str) -> int:
    if character == "":
        return ch2ix.get("<PAD>")
    elif character in string.punctuation:
        return ch2ix.get("<PUNC>")

    return ch2ix.get(character, ch2ix.get("<UNK>"))


def step_remove_tags(txt: str) -> str:
    return re.sub(r"<\/?[A-Z]+>", "", txt)


def step_thai_digit_to_arabic_digit(txt: str) -> str:
    return thai_digit_to_arabic_digit(txt)


def step_number_tag(txt: str, tag: str = "ttNumber") -> str:
    return re.sub(r"[0-9,]+", tag, txt)


def step_english_tag(txt: str, tag: str = "ttEnglish") -> str:
    return re.sub(r"[A-Za-z]+", tag, txt)


def step_new_line_as_space(txt: str) -> str:
    return re.sub(r"\n", " ", txt)


def step_remove_first_pipe(txt: str) -> str:
    return re.sub(r"^\|", "", txt)


def step_remove_last_pipe(txt: str) -> str:
    return re.sub(r"\|$", "", txt)


def preprocess(txt: str, steps=DEFAULT_PREPROCESSING_STEPS) -> str:
    for s in steps:
        if isinstance(s, str):
            txt = globals()["step_%s" % s](txt)
        elif callable(s):
            txt = s(txt)
    return txt


def expand_camel_case_to_tokens(w, verbose=0):
    if verbose:
        print(w)

    chars = ""
    for i, c in enumerate(w[1:]):
        ss = w[i-1:i+2]
        if verbose > 3:
            print(ss, chars)
        mm = CAMEL_CASE_RX.match(ss)
        if mm:
            chars = chars + "~" + mm[2]
            if verbose > 3:
                print(">>> match", mm[1], mm[2])
                print("******* ", chars)
        else:
            chars = chars + w[i]
    chars = chars + w[-1]
    return chars.split("~")


def find_words_from_preds(tokens, preds) -> List[str]:
    # Construct words from prediction labels {0, 1}
    curr_word = tokens[0]
    words = []
    for s, p in zip(tokens[1:], preds[1:]):
        if p == 0:
            curr_word = curr_word + s
        else:
            words.append(curr_word)
            curr_word = s

    words.append(curr_word)

    return words


def syllable_tokenize(txt: str) -> List[str]:
    # Proxy function for syllable tokenization, in case we want to try
    # a different syllable tokenizer.
    seps = txt.split(" ")

    new_tokens = []

    for i, s in enumerate(seps):
        tokens = ssg.syllable_tokenize(s)
        new_tokens.extend(tokens)

        if i < len(seps) - 1:
            new_tokens.append(" ")

    return new_tokens



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="attacut",
    version="1.0.4",
    description="Fast and Reasonably Accurate Word Tokenizer for Thai",
    url="https://github.com/PyThaiNLP/attacut",
    author="Pattarawat Chormai et al.",
    author_email="foomail@foo.com",
    packages=["attacut", "attacut.models", "attacut.artifacts"],
    install_requires=[
        "docopt>=0.6.2",
        "fire>=0.1.3",
        "nptyping>=0.2.0",
        "numpy>=1.17.0",
        "pyyaml>=5.1.2",
        "six>=1.12.0",
        "ssg>=0.0.4",
        "torch>=1.2.0",
    ],
    scripts=["scripts/attacut-cli"],
    package_data={"attacut": ["artifacts/**/*"]},
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Natural Language :: Thai",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import json
import os
import time
from typing import Callable, Dict, NamedTuple, Union

import yaml

from attacut import logger

log = logger.get_logger(__name__)


class ModelParams(NamedTuple):
    name: str
    params: str


class Timer:
    def __init__(self, name: str):
        self.name = name

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        self.stop = time.time()
        diff = self.stop - self.start
        log.info("Finished block: %s with %d seconds" % (self.name, diff))


def maybe(cond: bool, func: Callable[[], None], desc: str, verbose=0):
    if cond:
        func()
        return True
    else:
        log.info("Skipping: %s" % desc)


def maybe_create_dir(path: str):
    return maybe(
        not os.path.exists(path), lambda: os.mkdir(path), "create dir %s" % path
    )


def add_suffix_to_file_path(path: str, suffix: str) -> str:
    components = path.split("/")
    filename, ext = components[-1].split(".")

    components[-1] = "%s-%s.%s" % (filename, suffix, ext)

    return "/".join(components)


def wc_l(path: str) -> int:
    # count total lines in a file
    s = 0
    with open(path, "r") as f:
        for _ in f:
            s += 1
    return s


def save_training_params(dir_path: str, params: ModelParams):

    dir_path = "%s/params.yml" % dir_path
    print("Saving training params to %s" % dir_path)

    params = dict(params._asdict())

    with open(dir_path, "w") as outfile:
        yaml.dump(params, outfile, default_flow_style=False)


def load_training_params(path: str) -> ModelParams:
    with open("%s/params.yml" % path, "r") as f:
        params = yaml.load(f, Loader=yaml.BaseLoader)
        return ModelParams(**params)


def parse_model_params(ss: str) -> Dict[str, Union[int, float]]:
    params = dict()
    for pg in ss.split("|"):
        k, v = pg.split(":")

        if "." in v:
            params[k] = float(v)
        else:
            params[k] = int(v)
    return params


def load_dict(data_path: str) -> Dict:
    with open(data_path, "r", encoding="utf-8") as f:
        dd = json.load(f)

    log.info("loaded %d items from dict:%s" % (len(dd), data_path))

    return dd


def create_start_stop_indices(seq_lengths):
    # for old models, we might use it later???
    st_indices, sp_indices = [0], [seq_lengths[0]]

    for i, s in enumerate(seq_lengths[1:]):
        prev = sp_indices[i]
        st_indices.append(prev)
        sp_indices.append(prev + s)

    return list(zip(st_indices, sp_indices))



# Code below is taken from PyThaiNLP

_thai_arabic = {
    "๐": "0",
    "๑": "1",
    "๒": "2",
    "๓": "3",
    "๔": "4",
    "๕": "5",
    "๖": "6",
    "๗": "7",
    "๘": "8",
    "๙": "9",
}


def thai_digit_to_arabic_digit(text: str) -> str:
    """
    This function convert Thai digits (i.e. ๑, ๓, ๑๐) to Arabic digits
    (i.e. 1, 3, 10).
    :param str text: Text with Thai digits such as '๑', '๒', '๓'
    :return: Text with Thai digits being converted to Arabic digits
             such as '1', '2', '3'
    :rtype: str
    :Example:
    >>> from pythainlp.util import thai_digit_to_arabic_digit
    >>>
    >>> text = 'เป็นจำนวน ๑๒๓,๔๐๐.๒๕ บาท'
    >>> thai_digit_to_arabic_digit(text)
    เป็นจำนวน 123,400.25 บาท
    """
    if not text or not isinstance(text, str):
        return ""

    newtext = []
    for ch in text:
        if ch in _thai_arabic:
            newtext.append(_thai_arabic[ch])
        else:
            newtext.append(ch)

    return "".join(newtext)
